package com.sleepycat.je.log;

import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.ByteBuffer;

import com.sleepycat.je.DatabaseException;
import com.sleepycat.je.utilint.DbLsn;

public privileged aspect OutOfDiscHandler {

	/*
	 * Buffer may be rewritten because an IOException previously occurred.
	 */
	private boolean LogBuffer.rewriteAllowed = false;

	boolean LogBuffer.getRewriteAllowed() {
		return rewriteAllowed;
	}

	void LogBuffer.setRewriteAllowed() {
		rewriteAllowed = true;
	}

	void around(FileManager fm, LogBuffer fullBuffer) throws DatabaseException :
		execution(void FileManager.writeLogBuffer(LogBuffer)) && args(fullBuffer) && this(fm) {

		long firstLsn = fullBuffer.getFirstLsn();
		RandomAccessFile file = fm.endOfLog.getWritableFile(DbLsn
				.getFileNumber(firstLsn));
		ByteBuffer data = fullBuffer.getDataBuffer();

		try {
			try {
				/*
				 * Check that we do not overwrite unless the file only contains
				 * a header [#11915] [#12616].
				 */
				assert fullBuffer.getRewriteAllowed()
						|| (DbLsn.getFileOffset(firstLsn) >= file.length() || file
								.length() == FileManager.firstLogEntryOffset()) : "FileManager would overwrite non-empty file 0x"
					+ Long.toHexString(DbLsn.getFileNumber(firstLsn))
					+ " lsnOffset=0x"
					+ Long.toHexString(DbLsn.getFileOffset(firstLsn))
					+ " fileLength=0x" + Long.toHexString(file.length());

				proceed(fm, fullBuffer);
			} catch (DatabaseException dbe) {
				if (dbe.getCause() instanceof IOException) {
					throw (IOException)dbe.getCause();
				} else
					throw dbe;
			}
		} catch (IOException IOE) {

			/*
			 * Possibly an out of disk exception, but java.io will only tell us
			 * IOException with no indication of whether it's out of disk or
			 * something else.
			 * 
			 * Since we can't tell what sectors were actually written to disk,
			 * we need to change any commit records that might have made it out
			 * to disk to abort records. If they made it to disk on the write,
			 * then rewriting should allow them to be rewritten. See [11271].
			 */
			try {
				if (FileManager.IO_EXCEPTION_TESTING) {
					throw new IOException("generated for testing");
				}
				fm.writeToFile(file, data, DbLsn.getFileOffset(firstLsn));
			} catch (IOException IOE2) {
				fullBuffer.setRewriteAllowed();
				throw new DatabaseException(IOE2);
			}
		}
	}

}
